from __future__ import annotations

from datetime import datetime
import math
import os
from pathlib import Path
import sys
import tempfile
import time

from _common.types import Vector3Like


class Utils:
    @staticmethod
    def bFileExists(path: str | Path) -> bool:
        return Path(path).is_file()

    @staticmethod
    def bFolderExists(path: str | Path) -> bool:
        return Path(path).is_dir()

    @staticmethod
    def strGetFullPath(path: str | Path) -> str:
        return str(Path(path).resolve())

    @staticmethod
    def strGetFileName(path: str | Path) -> str:
        return Path(path).name

    @staticmethod
    def strGetFileNameWithoutExtension(path: str | Path) -> str:
        return Path(path).stem

    @staticmethod
    def strGetFolderName(path: str | Path) -> str:
        return Path(path).parent.name

    @staticmethod
    def strGetFileExtension(path: str | Path) -> str:
        return Path(path).suffix

    @staticmethod
    def strGetFileNameWithOtherExtension(path: str | Path, extension: str) -> str:
        p = Path(path)
        if extension and not extension.startswith("."):
            extension = "." + extension
        return str(p.with_suffix(extension))

    @staticmethod
    def strReplaceFileName(path: str | Path, new_name: str) -> str:
        return str(Path(path).with_name(new_name))

    @staticmethod
    def strShorten(text: str, max_len: int = 48) -> str:
        if max_len < 6:
            max_len = 6
        if len(text) <= max_len:
            return text
        lead = max_len // 2 - 1
        trail = max_len - lead - 3
        return text[:lead] + "..." + text[-trail:]

    @staticmethod
    def strStripQuotesFromPath(strPath: str) -> str:
        if len(strPath) >= 2 and strPath[0] == '"' and strPath[-1] == '"':
            return strPath[1:-1]
        return strPath

    @staticmethod
    def bWaitForFileExistence(strFile: str | Path, fTimeOut: float = 1000000.0) -> bool:
        deadline = time.monotonic() + max(0.0, float(fTimeOut))
        target = Path(strFile)
        while time.monotonic() < deadline:
            if target.exists():
                return True
            time.sleep(0.1)
        return target.exists()

    @staticmethod
    def strHomeFolder() -> str:
        return str(Path.home())

    @staticmethod
    def strDocumentsFolder() -> str:
        docs = Path.home() / "Documents"
        return str(docs if docs.exists() else Path.home())

    @staticmethod
    def strProjectRootFolder() -> str:
        return str(Path.cwd())

    @staticmethod
    def strPicoGKSourceCodeFolder() -> str:
        return str(Path(Utils.strProjectRootFolder()) / "PicoGK")

    @staticmethod
    def strExecutableFolder() -> str:
        return str(Path(sys.argv[0]).resolve().parent)

    @staticmethod
    def strDateTimeFilename(strPrefix: str, strPostfix: str) -> str:
        return f"{strPrefix}{datetime.now().strftime('%Y%m%d_%H%M%S')}{strPostfix}"

    @staticmethod
    def SetMatrixRow(mat: list[float], n: int, f1: float, f2: float, f3: float, f4: float) -> None:
        if len(mat) < 16:
            raise ValueError("mat must have at least 16 entries")
        if n < 0 or n > 3:
            raise ValueError("matrix row index must be in 0..3")
        base = n * 4
        mat[base] = float(f1)
        mat[base + 1] = float(f2)
        mat[base + 2] = float(f3)
        mat[base + 3] = float(f4)

    @staticmethod
    def matLookAt(vecEye: Vector3Like, vecLookAt: Vector3Like) -> tuple[float, ...]:
        ex, ey, ez = float(vecEye[0]), float(vecEye[1]), float(vecEye[2])
        lx, ly, lz = float(vecLookAt[0]), float(vecLookAt[1]), float(vecLookAt[2])

        vx, vy, vz = ex - lx, ey - ly, ez - lz
        vlen = math.sqrt(vx * vx + vy * vy + vz * vz)
        if vlen <= 1e-12:
            raise ValueError("vecEye must differ from vecLookAt")
        vx, vy, vz = vx / vlen, vy / vlen, vz / vlen

        zx, zy, zz = 0.0, 0.0, 1.0
        rx = zy * vz - zz * vy
        ry = zz * vx - zx * vz
        rz = zx * vy - zy * vx
        rlen = math.sqrt(rx * rx + ry * ry + rz * rz)
        if rlen <= 1e-12:
            rx, ry, rz = 1.0, 0.0, 0.0
            rlen = 1.0
        rx, ry, rz = rx / rlen, ry / rlen, rz / rlen

        ux = vy * rz - vz * ry
        uy = vz * rx - vx * rz
        uz = vx * ry - vy * rx

        mat = [0.0] * 16
        Utils.SetMatrixRow(mat, 0, rx, ux, vx, 0.0)
        Utils.SetMatrixRow(mat, 1, ry, uy, vy, 0.0)
        Utils.SetMatrixRow(mat, 2, rz, uz, vz, 0.0)
        Utils.SetMatrixRow(mat, 3, -(rx * ex + ry * ey + rz * ez), -(ux * ex + uy * ey + uz * ez), -(vx * ex + vy * ey + vz * ez), 1.0)
        return tuple(mat)

    @staticmethod
    def _voxel_size() -> float:
        from .library import Library

        return max(1e-6, float(getattr(Library, "voxel_size_mm", 0.5)))

    @staticmethod
    def mshCreateCube(
        vecScaleOrBox: Vector3Like | tuple[Vector3Like, Vector3Like] | None = None,
        vecOffsetMM: Vector3Like | None = None,
    ):
        from .mesh import Mesh

        if vecScaleOrBox is not None and len(vecScaleOrBox) == 2 and len(vecScaleOrBox[0]) == 3 and len(vecScaleOrBox[1]) == 3:  # type: ignore[index]
            bmin = vecScaleOrBox[0]  # type: ignore[index]
            bmax = vecScaleOrBox[1]  # type: ignore[index]
            scale = (float(bmax[0]) - float(bmin[0]), float(bmax[1]) - float(bmin[1]), float(bmax[2]) - float(bmin[2]))
            offset = (
                0.5 * (float(bmax[0]) + float(bmin[0])),
                0.5 * (float(bmax[1]) + float(bmin[1])),
                0.5 * (float(bmax[2]) + float(bmin[2])),
            )
        else:
            scale = tuple(float(v) for v in (vecScaleOrBox or (1.0, 1.0, 1.0)))
            offset = tuple(float(v) for v in (vecOffsetMM or (0.0, 0.0, 0.0)))

        sx, sy, sz = scale
        ox, oy, oz = offset

        mesh = Mesh()
        cube = [
            (-0.5 * sx + ox, -0.5 * sy + oy, -0.5 * sz + oz),
            (-0.5 * sx + ox, -0.5 * sy + oy, 0.5 * sz + oz),
            (-0.5 * sx + ox, 0.5 * sy + oy, -0.5 * sz + oz),
            (-0.5 * sx + ox, 0.5 * sy + oy, 0.5 * sz + oz),
            (0.5 * sx + ox, -0.5 * sy + oy, -0.5 * sz + oz),
            (0.5 * sx + ox, -0.5 * sy + oy, 0.5 * sz + oz),
            (0.5 * sx + ox, 0.5 * sy + oy, -0.5 * sz + oz),
            (0.5 * sx + ox, 0.5 * sy + oy, 0.5 * sz + oz),
        ]

        mesh.nAddTriangle(cube[0], cube[1], cube[3])
        mesh.nAddTriangle(cube[0], cube[3], cube[2])
        mesh.nAddTriangle(cube[4], cube[6], cube[7])
        mesh.nAddTriangle(cube[4], cube[7], cube[5])
        mesh.nAddTriangle(cube[0], cube[2], cube[6])
        mesh.nAddTriangle(cube[0], cube[6], cube[4])
        mesh.nAddTriangle(cube[1], cube[5], cube[7])
        mesh.nAddTriangle(cube[1], cube[7], cube[3])
        mesh.nAddTriangle(cube[2], cube[3], cube[7])
        mesh.nAddTriangle(cube[2], cube[7], cube[6])
        mesh.nAddTriangle(cube[0], cube[4], cube[5])
        mesh.nAddTriangle(cube[0], cube[5], cube[1])

        return mesh

    @staticmethod
    def mshCreateCylinder(vecScale: Vector3Like | None = None, vecOffsetMM: Vector3Like | None = None, iSides: int = 0):
        from .mesh import Mesh

        sx, sy, sz = tuple(float(v) for v in (vecScale or (1.0, 1.0, 1.0)))
        ox, oy, oz = tuple(float(v) for v in (vecOffsetMM or (0.0, 0.0, 0.0)))
        a = sx * 0.5
        b = sy * 0.5
        if iSides <= 0:
            av = a / Utils._voxel_size()
            bv = b / Utils._voxel_size()
            p = math.pi * (3.0 * (av + bv) - math.sqrt((3.0 * av + bv) * (av + 3.0 * bv)))
            iSides = max(3, 2 * int(math.ceil(p)))
        iSides = max(3, int(iSides))

        mesh = Mesh()
        bottom_center = (ox, oy, oz - sz * 0.5)
        top_center = (bottom_center[0], bottom_center[1], bottom_center[2] + sz)
        prev_bottom = (a + bottom_center[0], bottom_center[1], bottom_center[2])
        prev_top = (prev_bottom[0], prev_bottom[1], prev_bottom[2] + sz)
        step = 2.0 * math.pi / iSides
        for i in range(1, iSides + 1):
            angle = i * step
            this_bottom = (math.cos(angle) * a + bottom_center[0], math.sin(angle) * b + bottom_center[1], bottom_center[2])
            this_top = (this_bottom[0], this_bottom[1], this_bottom[2] + sz)

            mesh.nAddTriangle(top_center, prev_top, this_top)
            mesh.nAddTriangle(prev_bottom, this_bottom, prev_top)
            mesh.nAddTriangle(this_bottom, this_top, prev_top)
            mesh.nAddTriangle(bottom_center, this_bottom, prev_bottom)

            prev_bottom = this_bottom
            prev_top = this_top
        return mesh

    @staticmethod
    def mshCreateCone(vecScale: Vector3Like | None = None, vecOffsetMM: Vector3Like | None = None, iSides: int = 0):
        from .mesh import Mesh

        sx, sy, sz = tuple(float(v) for v in (vecScale or (1.0, 1.0, 1.0)))
        ox, oy, oz = tuple(float(v) for v in (vecOffsetMM or (0.0, 0.0, 0.0)))
        a = sx * 0.5
        b = sy * 0.5
        if iSides <= 0:
            av = a / Utils._voxel_size()
            bv = b / Utils._voxel_size()
            p = math.pi * (3.0 * (av + bv) - math.sqrt((3.0 * av + bv) * (av + 3.0 * bv)))
            iSides = max(3, 2 * int(math.ceil(p)))
        iSides = max(3, int(iSides))

        mesh = Mesh()
        bottom_center = (ox, oy, oz - sz * 0.5)
        top = (bottom_center[0], bottom_center[1], bottom_center[2] + sz)
        prev_bottom = (a + bottom_center[0], bottom_center[1], bottom_center[2])
        step = 2.0 * math.pi / iSides
        for i in range(1, iSides + 1):
            angle = i * step
            this_bottom = (math.cos(angle) * a + bottom_center[0], math.sin(angle) * b + bottom_center[1], bottom_center[2])
            mesh.nAddTriangle(prev_bottom, this_bottom, top)
            mesh.nAddTriangle(bottom_center, this_bottom, prev_bottom)
            prev_bottom = this_bottom
        return mesh

    @staticmethod
    def mshCreateGeoSphere(vecScale: Vector3Like | None = None, vecOffsetMM: Vector3Like | None = None, iSubdivisions: int = 0):
        from .mesh import Mesh

        sx, sy, sz = tuple(float(v) for v in (vecScale or (1.0, 1.0, 1.0)))
        ox, oy, oz = tuple(float(v) for v in (vecOffsetMM or (0.0, 0.0, 0.0)))
        rx, ry, rz = sx * 0.5, sy * 0.5, sz * 0.5

        def normalize_to_radii(v: tuple[float, float, float]) -> tuple[float, float, float]:
            x, y, z = v
            l = math.sqrt((x / max(1e-12, rx)) ** 2 + (y / max(1e-12, ry)) ** 2 + (z / max(1e-12, rz)) ** 2)
            if l <= 1e-12:
                return (0.0, 0.0, rz)
            return (x / l, y / l, z / l)

        t = (1.0 + math.sqrt(5.0)) / 2.0
        base = [
            (-1, t, 0), (1, t, 0), (-1, -t, 0), (1, -t, 0),
            (0, -1, t), (0, 1, t), (0, -1, -t), (0, 1, -t),
            (t, 0, -1), (t, 0, 1), (-t, 0, -1), (-t, 0, 1),
        ]
        verts = [normalize_to_radii((float(x) * rx, float(y) * ry, float(z) * rz)) for x, y, z in base]
        faces = [
            (0, 11, 5), (0, 5, 1), (0, 1, 7), (0, 7, 10), (0, 10, 11),
            (1, 5, 9), (5, 11, 4), (11, 10, 2), (10, 7, 6), (7, 1, 8),
            (3, 9, 4), (3, 4, 2), (3, 2, 6), (3, 6, 8), (3, 8, 9),
            (4, 9, 5), (2, 4, 11), (6, 2, 10), (8, 6, 7), (9, 8, 1),
        ]

        if iSubdivisions <= 0:
            area = 4.0 * math.pi * (((rx * ry) ** 1.6 + (ry * rz) ** 1.6 + (rz * rx) ** 1.6) / 3.0) ** (1.0 / 1.6)
            target = int(math.ceil(area / Utils._voxel_size() / Utils._voxel_size()))
            iSubdivisions = 1
            tris = 80
            while iSubdivisions < 8 and tris < target:
                iSubdivisions += 1
                tris = 20 * (1 << (2 * iSubdivisions))

        def midpoint(a: tuple[float, float, float], b: tuple[float, float, float]) -> tuple[float, float, float]:
            return normalize_to_radii(((a[0] + b[0]) * 0.5, (a[1] + b[1]) * 0.5, (a[2] + b[2]) * 0.5))

        for _ in range(max(0, int(iSubdivisions))):
            new_faces: list[tuple[int, int, int]] = []
            cache: dict[tuple[int, int], int] = {}

            def mid_idx(i: int, j: int) -> int:
                key = (i, j) if i < j else (j, i)
                if key in cache:
                    return cache[key]
                verts.append(midpoint(verts[i], verts[j]))
                idx = len(verts) - 1
                cache[key] = idx
                return idx

            for i0, i1, i2 in faces:
                a = mid_idx(i0, i1)
                b = mid_idx(i1, i2)
                c = mid_idx(i2, i0)
                new_faces.extend([(i0, a, c), (i1, b, a), (i2, c, b), (a, b, c)])
            faces = new_faces

        mesh = Mesh()
        for i0, i1, i2 in faces:
            a = verts[i0]
            b = verts[i1]
            c = verts[i2]
            mesh.nAddTriangle((a[0] + ox, a[1] + oy, a[2] + oz), (b[0] + ox, b[1] + oy, b[2] + oz), (c[0] + ox, c[1] + oy, c[2] + oz))
        return mesh

    def Dispose(self) -> None:
        return None


class TempFolder:
    def __init__(self) -> None:
        self.strFolder = tempfile.mkdtemp(prefix="picogk_")
        self._disposed = False

    def dispose(self) -> None:
        if self._disposed:
            return
        try:
            for file_path in Path(self.strFolder).glob("*"):
                if file_path.is_file():
                    file_path.unlink(missing_ok=True)
            Path(self.strFolder).rmdir()
        except Exception:
            pass
        self._disposed = True

    def __enter__(self) -> "TempFolder":
        return self

    def __exit__(self, exc_type: object, exc: object, tb: object) -> None:
        self.dispose()




from __future__ import annotations

import numpy as np
from .control_point_spline import ControlPointSpline
from .i_spline import ISpline

from .._types import Vec3, Vector3Like, as_np3, as_vec3, normalized
from ..frames.local_frames import LocalFrame


class TangentialControlSpline(ISpline):
    def __init__(self, oBSpline: ControlPointSpline) -> None:
        self.m_oBSpline = oBSpline

    @classmethod
    def from_vectors(
        cls,
        vecStart: Vector3Like,
        vecEnd: Vector3Like,
        vecStartDir: Vector3Like,
        vecEndDir: Vector3Like,
        fStartTangentStrength: float = -1.0,
        fEndTangentStrenth: float = -1.0,
        bRelativeStartStrength: bool = False,
        bRelativeEndStrength: bool = False,
    ) -> "TangentialControlSpline":
        start = as_np3(vecStart)
        end = as_np3(vecEnd)
        start_dir = normalized(vecStartDir)
        end_dir = normalized(vecEndDir)
        distance = float(np.linalg.norm(start - end))

        start_strength = fStartTangentStrength
        end_strength = fEndTangentStrenth

        if start_strength == -1.0:
            start_strength = 0.3 * distance
        if end_strength == -1.0:
            end_strength = 0.3 * distance
        if bRelativeStartStrength:
            start_strength *= distance
        if bRelativeEndStrength:
            end_strength *= distance

        control_points = [
            as_vec3(start),
            as_vec3(start + start_strength * start_dir),
            as_vec3(end - end_strength * end_dir),
            as_vec3(end),
        ]
        return cls(ControlPointSpline(control_points))

    @classmethod
    def from_frames(
        cls,
        oStartFrame: LocalFrame,
        oEndFrame: LocalFrame,
        fStartTangentStrength: float = -1.0,
        fEndTangentStrenth: float = -1.0,
        bRelativeStartStrength: bool = False,
        bRelativeEndStrength: bool = False,
    ) -> "TangentialControlSpline":
        return cls.from_vectors(
            oStartFrame.vecGetPosition(),
            oEndFrame.vecGetPosition(),
            oStartFrame.vecGetLocalZ(),
            oEndFrame.vecGetLocalZ(),
            fStartTangentStrength,
            fEndTangentStrenth,
            bRelativeStartStrength,
            bRelativeEndStrength,
        )

    def aGetPoints(self, nSamples: int | None = 500) -> list[Vec3]:
        if nSamples is None:
            nSamples = 500
        return self.m_oBSpline.aGetPoints(nSamples)




from __future__ import annotations

import math
from typing import TYPE_CHECKING, Sequence

import numpy as np

if TYPE_CHECKING:
    from .vedo_viewer import VedoViewer

class AnimGroupMatrixRotate:
    def __init__(
        self,
        oViewer: "VedoViewer",
        nGroup: int,
        matInit: Sequence[float],
        vecAxis: Sequence[float],
        fDegrees: float,
    ) -> None:
        if len(matInit) != 16:
            raise ValueError("matInit must contain 16 values")
        self._viewer = oViewer
        self._group = int(nGroup)
        self._mat_init = np.array(matInit, dtype=np.float64).reshape((4, 4))
        axis = np.array(vecAxis, dtype=np.float64)
        n = float(np.linalg.norm(axis))
        if n <= 1e-12:
            raise ValueError("vecAxis must be non-zero")
        self._axis = axis / n
        self._degrees = float(fDegrees)

    def Do(self, fFactor: float) -> None:
        theta = math.radians(float(fFactor) * self._degrees)
        x, y, z = self._axis.tolist()
        c = math.cos(theta)
        s = math.sin(theta)
        t = 1.0 - c
        rot = np.array(
            [
                [t * x * x + c, t * x * y - s * z, t * x * z + s * y, 0.0],
                [t * x * y + s * z, t * y * y + c, t * y * z - s * x, 0.0],
                [t * x * z - s * y, t * y * z + s * x, t * z * z + c, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            dtype=np.float64,
        )
        mat = self._mat_init @ rot
        self._viewer.SetGroupMatrix(self._group, tuple(float(v) for v in mat.reshape(-1)))


class AnimViewRotate:
    def __init__(self, oViewer: "VedoViewer", vecFrom: Sequence[float], vecTo: Sequence[float]) -> None:
        if len(vecFrom) < 2 or len(vecTo) < 2:
            raise ValueError("vecFrom and vecTo must contain at least 2 values")
        self._viewer = oViewer
        self._from = (float(vecFrom[0]), float(vecFrom[1]))
        self._to = (float(vecTo[0]), float(vecTo[1]))

    def Do(self, fFactor: float) -> None:
        factor = float(fFactor)
        orbit = self._from[0] + ((self._to[0] - self._from[0]) * factor)
        elevation = self._from[1] + ((self._to[1] - self._from[1]) * factor)
        self._viewer.SetViewAngles(orbit, elevation)




from __future__ import annotations

import ctypes
from enum import IntEnum
from typing import TYPE_CHECKING, cast

from _common.types import Vector3Like

from .._base import HandleOwner
from .._errors import PicoGKInvalidHandleError
from .._types import Vec3, as_vec3, vec3_tuple

from .library import Library

if TYPE_CHECKING:
    from .fields import ScalarField, VectorField
    from .._core.voxels import Voxels

class FieldMetadata(HandleOwner):
    class Type(IntEnum):
        UNKNOWN = -1
        STRING = 0
        FLOAT = 1
        VECTOR = 2

    @classmethod
    def from_voxels(cls, voxels: "Voxels") -> "FieldMetadata":
        h = Library._lib().Metadata_hFromVoxels(voxels.handle)
        if not h:
            raise PicoGKInvalidHandleError("Metadata_hFromVoxels returned null")
        return cls(h)

    @classmethod
    def from_scalar_field(cls, field: "ScalarField") -> "FieldMetadata":
        h = Library._lib().Metadata_hFromScalarField(field.handle)
        if not h:
            raise PicoGKInvalidHandleError("Metadata_hFromScalarField returned null")
        return cls(h)

    @classmethod
    def from_vector_field(cls, field: "VectorField") -> "FieldMetadata":
        h = Library._lib().Metadata_hFromVectorField(field.handle)
        if not h:
            raise PicoGKInvalidHandleError("Metadata_hFromVectorField returned null")
        return cls(h)

    def _destroy_native(self) -> None:
        Library._lib().Metadata_Destroy(self._handle)

    def count(self) -> int:
        self._ensure_open()
        return int(Library._lib().Metadata_nCount(self.handle))

    nCount = count

    def name_at(self, index: int) -> str:
        self._ensure_open()
        n = int(Library._lib().Metadata_nNameLengthAt(self.handle, int(index))) + 1
        buf = ctypes.create_string_buffer(n)
        ok = bool(Library._lib().Metadata_bGetNameAt(self.handle, int(index), buf, int(n)))
        if not ok:
            raise IndexError(index)
        return buf.value.decode("utf-8")

    def bGetNameAt(self, index: int) -> tuple[bool, str]:
        try:
            return True, self.name_at(index)
        except IndexError:
            return False, ""

    def type_at(self, name: str) -> "FieldMetadata.Type":
        self._ensure_open()
        return FieldMetadata.Type(int(Library._lib().Metadata_nTypeAt(self.handle, name.encode("utf-8"))))

    eTypeAt = type_at

    @classmethod
    def strTypeName(cls, value_type: "FieldMetadata.Type") -> str:
        mapping = {
            cls.Type.UNKNOWN: "unknown",
            cls.Type.STRING: "string",
            cls.Type.FLOAT: "float",
            cls.Type.VECTOR: "vector",
        }
        return mapping.get(value_type, "undefined")

    def strTypeAt(self, name: str) -> str:
        return self.strTypeName(self.type_at(name))

    def get_string(self, name: str) -> str:
        self._ensure_open()
        name_b = name.encode("utf-8")
        n = int(Library._lib().Metadata_nStringLengthAt(self.handle, name_b)) + 1
        buf = ctypes.create_string_buffer(n)
        ok = bool(Library._lib().Metadata_bGetStringAt(self.handle, name_b, buf, int(n)))
        if not ok:
            raise KeyError(name)
        return buf.value.decode("utf-8")

    def bGetValueAt(self, name: str) -> tuple[bool, object]:
        value_type = self.type_at(name)
        try:
            if value_type == self.Type.STRING:
                return True, self.get_string(name)
            if value_type == self.Type.FLOAT:
                return True, self.get_float(name)
            if value_type == self.Type.VECTOR:
                return True, self.get_vector(name)
        except KeyError:
            return False, ""
        return False, ""

    def get_float(self, name: str) -> float:
        self._ensure_open()
        out = ctypes.c_float()
        ok = bool(Library._lib().Metadata_bGetFloatAt(self.handle, name.encode("utf-8"), ctypes.byref(out)))
        if not ok:
            raise KeyError(name)
        return float(out.value)

    def get_vector(self, name: str) -> tuple[float, float, float]:
        self._ensure_open()
        out = Vec3()
        ok = bool(Library._lib().Metadata_bGetVectorAt(self.handle, name.encode("utf-8"), ctypes.byref(out)))
        if not ok:
            raise KeyError(name)
        return vec3_tuple(out)

    def set_string(self, name: str, value: str) -> "FieldMetadata":
        self._ensure_open()
        Library._lib().Metadata_SetStringValue(self.handle, name.encode("utf-8"), value.encode("utf-8"))
        return self

    def set_float(self, name: str, value: float) -> "FieldMetadata":
        self._ensure_open()
        Library._lib().Metadata_SetFloatValue(self.handle, name.encode("utf-8"), ctypes.c_float(value))
        return self

    def set_vector(self, name: str, value: Vector3Like) -> "FieldMetadata":
        self._ensure_open()
        v = as_vec3(value)
        Library._lib().Metadata_SetVectorValue(self.handle, name.encode("utf-8"), ctypes.byref(v))
        return self

    def SetValue(self, name: str, value: object) -> "FieldMetadata":
        if isinstance(value, str):
            return self.set_string(name, value)
        if isinstance(value, (int, float)):
            return self.set_float(name, float(value))
        return self.set_vector(name, cast(Vector3Like, value))

    def remove(self, name: str) -> "FieldMetadata":
        self._ensure_open()
        Library._lib().MetaData_RemoveValue(self.handle, name.encode("utf-8"))
        return self

    RemoveValue = remove

    def __str__(self) -> str:
        rows: list[str] = [f"Metadata table with {self.count()} items"]
        for index in range(self.count()):
            ok, name = self.bGetNameAt(index)
            if not ok:
                continue
            value_ok, value = self.bGetValueAt(name)
            rows.append(f"  {name} ({self.strTypeAt(name)}): {value if value_ok else ''}")
        return "\n".join(rows)

    def Dispose(self) -> None:
        self.close()






from __future__ import annotations

from picogk import Lattice

from .._helpers import _add_sampled_beam
from ..beam_thickness import IBeamThickness
from ..unit_cells import IUnitCell
from .i_lattice_type import ILatticeType


class BodyCentreLattice(ILatticeType):
	def AddCell(self, oLattice: Lattice, xCell: IUnitCell, xBeamThickness: IBeamThickness, nSubSamples: int = 2) -> None:
		aCornerPoints = xCell.aGetCornerPoints()
		if len(aCornerPoints) != 8:
			raise ValueError("Body Centre Lattice only supports Unit Cells with 8 Corners.")

		vecCPt = xCell.vecGetCellCentre()
		for vecCorner in aCornerPoints:
			_add_sampled_beam(oLattice, vecCorner, vecCPt, xBeamThickness, nSubSamples)


__all__ = ["BodyCentreLattice"]


from .icosahedral_face import IcosehedralFace
from .quasi_tile import QuasiTile
from .quasi_tile_01 import QuasiTile_01
from .quasi_tile_02 import QuasiTile_02
from .quasi_tile_03 import QuasiTile_03
from .quasi_tile_04 import QuasiTile_04
from .quasi_tile_inflation import QuasiTileInflation
from .quasi_crystal import QuasiCrystal

__all__ = [
    "IcosehedralFace",
    "QuasiTile",
    "QuasiTile_01",
    "QuasiTile_02",
    "QuasiTile_03",
    "QuasiTile_04",
    "QuasiTileInflation",
    "QuasiCrystal",
]



from __future__ import annotations

from .._types import as_np3
from ..utils import SplineOperations
from .color_scale_2d import IColorScale
from .rainbox_spectrum import ISpectrum


class ColorScale3D(IColorScale):
	def __init__(self, xSpectrum: ISpectrum, fMinValue: float, fMaxValue: float) -> None:
		self.m_fMinValue = float(fMinValue)
		self.m_fMaxValue = float(fMaxValue)
		smooth = SplineOperations.aGetNURBSpline(xSpectrum.aGetRawRGBList(), 500)
		self.m_aSmoothRGBList = [as_np3(p) for p in SplineOperations.aGetReparametrizedSpline(smooth, 500)]

	def fGetMinValue(self) -> float:
		return self.m_fMinValue

	def fGetMaxValue(self) -> float:
		return self.m_fMaxValue

	def clrGetColor(self, fValue: float) -> tuple[float, float, float, float]:
		value = min(max(float(fValue), self.m_fMinValue), self.m_fMaxValue)
		ratio = (value - self.m_fMinValue) / (self.m_fMaxValue - self.m_fMinValue)
		idx = int(ratio * (len(self.m_aSmoothRGBList) - 1))
		rgb = self.m_aSmoothRGBList[idx]
		return (float(rgb[0] / 255.0), float(rgb[1] / 255.0), float(rgb[2] / 255.0), 1.0)


__all__ = ["ColorScale3D"]



from .body_centre_lattice import BodyCentreLattice
from .i_lattice_type import ILatticeType
from .octahedron_lattice import OctahedronLattice
from .random_spline_lattice import RandomSplineLattice

__all__ = ["ILatticeType", "BodyCentreLattice", "OctahedronLattice", "RandomSplineLattice"]


from __future__ import annotations

import math

import numpy as np
from picogk import IImplicit

from .._types import Vector3Like, as_np3, as_vec3


class ImplicitGyroid(IImplicit):
    def __init__(self, fUnitSize: float, fThicknessRatio: float) -> None:
        self.m_fFrequencyScale = (2.0 * math.pi) / float(fUnitSize)
        self.m_fThicknessRatio = float(fThicknessRatio)

    @staticmethod
    def fGetThicknessRatio(fWallThickness: float, fUnitSize: float) -> float:
        return float(fWallThickness) * 10.0 / float(fUnitSize)

    def fSignedDistance(self, vecPt: Vector3Like) -> float:
        x, y, z = as_vec3(vecPt)
        dist = (
            math.sin(self.m_fFrequencyScale * x) * math.cos(self.m_fFrequencyScale * y)
            + math.sin(self.m_fFrequencyScale * y) * math.cos(self.m_fFrequencyScale * z)
            + math.sin(self.m_fFrequencyScale * z) * math.cos(self.m_fFrequencyScale * x)
        )
        return float(abs(dist) - 0.5 * self.m_fThicknessRatio)


class ImplicitSphere(IImplicit):
    def __init__(self, vecCentre: Vector3Like, fRadius: float) -> None:
        self.m_vecCentre = as_np3(vecCentre)
        self.m_fRadius = float(fRadius)

    def fSignedDistance(self, vecPt: Vector3Like) -> float:
        return float(np.linalg.norm(as_np3(vecPt) - self.m_vecCentre) - self.m_fRadius)


class ImplicitGenus(IImplicit):
    def __init__(self, fGap: float) -> None:
        self.m_fGap = float(fGap)

    def fSignedDistance(self, vecPt: Vector3Like) -> float:
        x, y, z = as_vec3(vecPt)
        return float(
            2.0 * y * (y * y - 3.0 * x * x) * (1.0 - z * z)
            + (x * x + y * y) ** 2
            - (9.0 * z * z - 1.0) * (1.0 - z * z)
            - self.m_fGap
        )


class ImplicitSuperEllipsoid(IImplicit):
    def __init__(
        self,
        vecCentre: Vector3Like,
        fAx: float,
        fAy: float,
        fAz: float,
        fEpsilon1: float,
        fEpsilon2: float,
    ) -> None:
        self.m_vecCentre = as_np3(vecCentre)
        self.m_fAx = float(fAx)
        self.m_fAy = float(fAy)
        self.m_fAz = float(fAz)
        self.m_fEpsilon1 = float(fEpsilon1)
        self.m_fEpsilon2 = float(fEpsilon2)

    def fSignedDistance(self, vecPt: Vector3Like) -> float:
        x, y, z = as_np3(vecPt)
        d_x = abs(x + self.m_vecCentre[0]) / self.m_fAx
        d_y = abs(y + self.m_vecCentre[1]) / self.m_fAy
        d_z = abs(z + self.m_vecCentre[2]) / self.m_fAz
        dist = (
            (d_x ** (2.0 / self.m_fEpsilon2) + d_y ** (2.0 / self.m_fEpsilon2))
            ** (self.m_fEpsilon2 / self.m_fEpsilon1)
            + d_z ** (2.0 / self.m_fEpsilon1)
        )
        return float(dist - 1.0)


__all__ = [
    "ImplicitGyroid",
    "ImplicitSphere",
    "ImplicitGenus",
    "ImplicitSuperEllipsoid",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from picogk import Voxels
from shape_kernel._types import Vector3Like

if TYPE_CHECKING:
	from ..unit_cells.i_unit_cell import IUnitCell


class IBeamThickness(ABC):
	@abstractmethod
	def fGetBeamThickness(self, vecPt: Vector3Like) -> float:
		raise NotImplementedError

	@abstractmethod
	def UpdateCell(self, xCell: IUnitCell) -> None:
		raise NotImplementedError

	@abstractmethod
	def SetBoundingVoxels(self, voxBounding: Voxels) -> None:
		raise NotImplementedError


__all__ = ["IBeamThickness"]


from __future__ import annotations

import ctypes
import math
from typing import TYPE_CHECKING, Any, Sequence, cast

import numpy as np

from _common.types import Vector3Like
from picogk._core.mesh_io import EStlUnit, MeshIo
from picogk._core.mesh_mat import MeshMath

from .._base import HandleOwner
from .._errors import PicoGKInvalidHandleError
from .._types import BBox3, Triangle, Vec3, as_vec3, vec3_tuple

from .library import Library

if TYPE_CHECKING:
    from .._core.voxels import Voxels

class Mesh(HandleOwner):
    def __init__(self, handle: int | ctypes.c_void_p | None = None, *, owns_handle: bool = True) -> None:
        if handle is None:
            handle = Library._lib().Mesh_hCreate()
        if not handle:
            raise PicoGKInvalidHandleError("Mesh_hCreate returned null")
        super().__init__(handle, owns_handle=owns_handle)
        # C# parity fields used by STL IO helpers.
        self.m_strLoadHeaderData: str = ""
        self.m_eLoadUnits: EStlUnit = EStlUnit.AUTO

    @classmethod
    def from_voxels(cls, voxels: "Voxels") -> "Mesh":
        h = Library._lib().Mesh_hCreateFromVoxels(voxels.handle)
        if not h:
            raise PicoGKInvalidHandleError("Mesh_hCreateFromVoxels returned null")
        return cls(h)

    def _destroy_native(self) -> None:
        Library._lib().Mesh_Destroy(self._handle)

    def is_valid(self) -> bool:
        return bool(Library._lib().Mesh_bIsValid(self.handle))

    def add_vertex(self, xyz: Vector3Like) -> int:
        self._ensure_open()
        v = as_vec3(xyz)
        return int(Library._lib().Mesh_nAddVertex(self.handle, ctypes.byref(v)))

    nAddVertex = add_vertex

    def vertex_count(self) -> int:
        self._ensure_open()
        return int(Library._lib().Mesh_nVertexCount(self.handle))

    nVertexCount = vertex_count

    def get_vertex(self, index: int) -> tuple[float, float, float]:
        self._ensure_open()
        v = Vec3()
        Library._lib().Mesh_GetVertex(self.handle, int(index), ctypes.byref(v))
        return vec3_tuple(v)

    vecVertexAt = get_vertex

    def add_vertices(self, vertices: Sequence[Vector3Like]) -> list[int]:
        return [self.add_vertex(vertex) for vertex in vertices]

    AddVertices = add_vertices

    def add_triangle_indices(self, a: int, b: int, c: int) -> int:
        self._ensure_open()
        tri = Triangle(int(a), int(b), int(c))
        return int(Library._lib().Mesh_nAddTriangle(self.handle, ctypes.byref(tri)))

    def add_triangle(self, a_xyz: Vector3Like, b_xyz: Vector3Like, c_xyz: Vector3Like) -> int:
        ai = self.add_vertex(a_xyz)
        bi = self.add_vertex(b_xyz)
        ci = self.add_vertex(c_xyz)
        return self.add_triangle_indices(ai, bi, ci)

    def nAddTriangle(self, *args: object) -> int:
        if len(args) == 1 and isinstance(args[0], Triangle):
            tri = cast(Triangle, args[0])
            return self.add_triangle_indices(tri.a, tri.b, tri.c)
        if len(args) == 3 and all(isinstance(arg, int) for arg in args):
            return self.add_triangle_indices(cast(int, args[0]), cast(int, args[1]), cast(int, args[2]))
        if len(args) == 3:
            return self.add_triangle(cast(Vector3Like, args[0]), cast(Vector3Like, args[1]), cast(Vector3Like, args[2]))
        raise TypeError("nAddTriangle expects a Triangle, 3 ints, or 3 xyz vectors")

    def triangle_count(self) -> int:
        self._ensure_open()
        return int(Library._lib().Mesh_nTriangleCount(self.handle))

    nTriangleCount = triangle_count

    def get_triangle(self, index: int) -> tuple[int, int, int]:
        self._ensure_open()
        tri = Triangle()
        Library._lib().Mesh_GetTriangle(self.handle, int(index), ctypes.byref(tri))
        return (int(tri.a), int(tri.b), int(tri.c))

    oTriangleAt = get_triangle

    def get_triangle_vertices(self, index: int) -> tuple[tuple[float, float, float], tuple[float, float, float], tuple[float, float, float]]:
        self._ensure_open()
        a = Vec3()
        b = Vec3()
        c = Vec3()
        Library._lib().Mesh_GetTriangleV(self.handle, int(index), ctypes.byref(a), ctypes.byref(b), ctypes.byref(c))
        return vec3_tuple(a), vec3_tuple(b), vec3_tuple(c)

    GetTriangle = get_triangle_vertices

    def bounding_box(self) -> tuple[tuple[float, float, float], tuple[float, float, float]]:
        self._ensure_open()
        box = BBox3()
        Library._lib().Mesh_GetBoundingBox(self.handle, ctypes.byref(box))
        return vec3_tuple(box.min), vec3_tuple(box.max)

    oBoundingBox = bounding_box

    def add_quad_indices(self, n0: int, n1: int, n2: int, n3: int, flipped: bool = False) -> "Mesh":
        if flipped:
            self.add_triangle_indices(n0, n2, n1)
            self.add_triangle_indices(n0, n3, n2)
        else:
            self.add_triangle_indices(n0, n1, n2)
            self.add_triangle_indices(n0, n2, n3)
        return self

    def add_quad(self, v0: Vector3Like, v1: Vector3Like, v2: Vector3Like, v3: Vector3Like, flipped: bool = False) -> "Mesh":
        n0 = self.add_vertex(v0)
        n1 = self.add_vertex(v1)
        n2 = self.add_vertex(v2)
        n3 = self.add_vertex(v3)
        return self.add_quad_indices(n0, n1, n2, n3, flipped)

    def AddQuad(self, *args: object, bFlipped: bool = False) -> "Mesh":
        if len(args) == 4 and all(isinstance(arg, int) for arg in args):
            return self.add_quad_indices(cast(int, args[0]), cast(int, args[1]), cast(int, args[2]), cast(int, args[3]), bFlipped)
        if len(args) == 4:
            return self.add_quad(cast(Vector3Like, args[0]), cast(Vector3Like, args[1]), cast(Vector3Like, args[2]), cast(Vector3Like, args[3]), bFlipped)
        raise TypeError("AddQuad expects 4 indices or 4 xyz vectors")

    def append(self, other: "Mesh") -> "Mesh":
        for index in range(other.triangle_count()):
            a, b, c = other.get_triangle_vertices(index)
            self.add_triangle(a, b, c)
        return self

    Append = append

    def create_transformed(self, scale: Vector3Like = (1.0, 1.0, 1.0), offset: Vector3Like = (0.0, 0.0, 0.0)) -> "Mesh":
        scale_v = scale
        offset_v = offset
        out = Mesh()
        for index in range(self.triangle_count()):
            a, b, c = self.get_triangle_vertices(index)
            out.add_triangle(
                (a[0] * scale_v[0] + offset_v[0], a[1] * scale_v[1] + offset_v[1], a[2] * scale_v[2] + offset_v[2]),
                (b[0] * scale_v[0] + offset_v[0], b[1] * scale_v[1] + offset_v[1], b[2] * scale_v[2] + offset_v[2]),
                (c[0] * scale_v[0] + offset_v[0], c[1] * scale_v[1] + offset_v[1], c[2] * scale_v[2] + offset_v[2]),
            )
        return out

    mshCreateTransformed = create_transformed

    def create_mirrored(self, plane_point: Vector3Like, plane_normal: Vector3Like) -> "Mesh":
        px, py, pz = plane_point
        nx, ny, nz = plane_normal
        length = math.sqrt(nx * nx + ny * ny + nz * nz)
        if length <= 1e-12:
            raise ValueError("plane_normal must be non-zero")
        nx /= length
        ny /= length
        nz /= length

        def mirror(v: tuple[float, float, float]) -> tuple[float, float, float]:
            vx, vy, vz = v
            dx = vx - px
            dy = vy - py
            dz = vz - pz
            dist = dx * nx + dy * ny + dz * nz
            return (vx - 2 * dist * nx, vy - 2 * dist * ny, vz - 2 * dist * nz)

        out = Mesh()
        for index in range(self.triangle_count()):
            a, b, c = self.get_triangle_vertices(index)
            out.add_triangle(mirror(a), mirror(b), mirror(c))
        return out

    mshCreateMirrored = create_mirrored

    def to_numpy(self) -> tuple[np.ndarray, np.ndarray]:
        self._ensure_open()
        vertices = np.zeros((self.vertex_count(), 3), dtype=np.float32)
        for i in range(vertices.shape[0]):
            vertices[i, :] = np.array(self.get_vertex(i), dtype=np.float32)

        triangles = np.zeros((self.triangle_count(), 3), dtype=np.int32)
        for i in range(triangles.shape[0]):
            triangles[i, :] = np.array(self.get_triangle(i), dtype=np.int32)

        return vertices, triangles

    @staticmethod
    def bPointLiesOnTriangle(vecP: Vector3Like, vecA: Vector3Like, vecB: Vector3Like, vecC: Vector3Like) -> bool:
        return MeshMath.bPointLiesOnTriangle(vecP, vecA, vecB, vecC)

    def bFindTriangleFromSurfacePoint(self, vecSurfacePoint: Vector3Like) -> tuple[bool, int]:
        return MeshMath.bFindTriangleFromSurfacePoint(self, vecSurfacePoint)

    def voxVoxelizeHollow(self, fThickness: float) -> "Voxels":
        from .triangle_voxelization import TriangleVoxelization

        return TriangleVoxelization.voxVoxelizeHollow(self, float(fThickness))

    @staticmethod
    def mshFromStlFile(
        strFilePath: str,
        eLoadUnit: Any = None,
        fPostScale: float = 1.0,
        vecPostOffsetMM: Vector3Like | None = None,
    ) -> "Mesh":

        load_unit = EStlUnit.AUTO if eLoadUnit is None else eLoadUnit
        return MeshIo.mshFromStlFile(strFilePath, cast(Any, load_unit), fPostScale, vecPostOffsetMM)

    def SaveToStlFile(
        self,
        strFilePath: str,
        eUnit: Any = None,
        vecOffsetMM: Vector3Like | None = None,
        fScale: float = 1.0,
    ) -> None:

        unit = EStlUnit.AUTO if eUnit is None else eUnit
        MeshIo.SaveToStlFile(self, strFilePath, cast(Any, unit), vecOffsetMM, fScale)




from __future__ import annotations

import ctypes
from dataclasses import dataclass
from typing import Any

from ._config import candidate_library_paths
from ._errors import PicoGKLoadError
from ._types import BBox3, ColorFloat, Mat4, Triangle, Vec2, Vec3


CallbackImplicitDistance = ctypes.CFUNCTYPE(ctypes.c_float, ctypes.POINTER(Vec3))
CallbackScalarFieldTraverse = ctypes.CFUNCTYPE(None, ctypes.POINTER(Vec3), ctypes.c_float)
CallbackVectorFieldTraverse = ctypes.CFUNCTYPE(None, ctypes.POINTER(Vec3), ctypes.POINTER(Vec3))

InfoCallback = ctypes.CFUNCTYPE(None, ctypes.c_char_p, ctypes.c_bool)
UpdateCallback = ctypes.CFUNCTYPE(
    None,
    ctypes.c_void_p,
    ctypes.POINTER(Vec2),
    ctypes.POINTER(ColorFloat),
    ctypes.POINTER(Mat4),
    ctypes.POINTER(Mat4),
    ctypes.POINTER(Mat4),
    ctypes.POINTER(Vec3),
    ctypes.POINTER(Vec3),
)
KeyPressedCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.c_int)
MouseMovedCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(Vec2))
MouseButtonCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int, ctypes.POINTER(Vec2))
ScrollWheelCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(Vec2), ctypes.POINTER(Vec2))
WindowSizeCallback = ctypes.CFUNCTYPE(None, ctypes.c_void_p, ctypes.POINTER(Vec2))


@dataclass
class Native:
    lib: ctypes.CDLL


def _bind(lib: ctypes.CDLL, name: str, restype: Any, argtypes: list[Any]) -> None:
    fn = getattr(lib, name)
    fn.restype = restype
    fn.argtypes = argtypes


def _bind_if_exists(lib: ctypes.CDLL, name: str, restype: Any, argtypes: list[Any]) -> None:
    if hasattr(lib, name):
        _bind(lib, name, restype, argtypes)


def _load_lib() -> ctypes.CDLL:
    errors: list[str] = []
    for candidate in candidate_library_paths():
        try:
            return ctypes.CDLL(candidate)
        except OSError as exc:
            errors.append(f"{candidate}: {exc}")
    msg = "Unable to load PicoGK runtime. Tried:\n" + "\n".join(errors)
    raise PicoGKLoadError(msg)


def _configure(lib: ctypes.CDLL) -> None:
    # Library
    _bind(lib, "Library_Init", None, [ctypes.c_float])
    _bind(lib, "Library_Destroy", None, [])
    _bind(lib, "Library_GetName", None, [ctypes.c_char_p])
    _bind(lib, "Library_GetVersion", None, [ctypes.c_char_p])
    _bind(lib, "Library_GetBuildInfo", None, [ctypes.c_char_p])
    _bind(lib, "Library_VoxelsToMm", None, [ctypes.POINTER(Vec3), ctypes.POINTER(Vec3)])
    _bind(lib, "Library_MmToVoxels", None, [ctypes.POINTER(Vec3), ctypes.POINTER(Vec3)])

    # Mesh
    _bind(lib, "Mesh_hCreate", ctypes.c_void_p, [])
    _bind(lib, "Mesh_hCreateFromVoxels", ctypes.c_void_p, [ctypes.c_void_p])
    _bind(lib, "Mesh_bIsValid", ctypes.c_bool, [ctypes.c_void_p])
    _bind(lib, "Mesh_Destroy", None, [ctypes.c_void_p])
    _bind(lib, "Mesh_nAddVertex", ctypes.c_int, [ctypes.c_void_p, ctypes.POINTER(Vec3)])
    _bind(lib, "Mesh_nVertexCount", ctypes.c_int, [ctypes.c_void_p])
    _bind(lib, "Mesh_GetVertex", None, [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(Vec3)])
    _bind(lib, "Mesh_nAddTriangle", ctypes.c_int, [ctypes.c_void_p, ctypes.POINTER(Triangle)])
    _bind(lib, "Mesh_nTriangleCount", ctypes.c_int, [ctypes.c_void_p])
    _bind(lib, "Mesh_GetTriangle", None, [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(Triangle)])
    _bind(lib, "Mesh_GetTriangleV", None, [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(Vec3), ctypes.POINTER(Vec3), ctypes.POINTER(Vec3)])
    _bind(lib, "Mesh_GetBoundingBox", None, [ctypes.c_void_p, ctypes.POINTER(BBox3)])

    # Lattice
    _bind(lib, "Lattice_hCreate", ctypes.c_void_p, [])
    _bind(lib, "Lattice_bIsValid", ctypes.c_bool, [ctypes.c_void_p])
    _bind(lib, "Lattice_Destroy", None, [ctypes.c_void_p])
    _bind(lib, "Lattice_AddSphere", None, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.c_float])
    _bind(lib, "Lattice_AddBeam", None, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.POINTER(Vec3), ctypes.c_float, ctypes.c_float, ctypes.c_bool])

    # Voxels
    _bind(lib, "Voxels_hCreate", ctypes.c_void_p, [])
    _bind(lib, "Voxels_hCreateCopy", ctypes.c_void_p, [ctypes.c_void_p])
    _bind(lib, "Voxels_bIsValid", ctypes.c_bool, [ctypes.c_void_p])
    _bind(lib, "Voxels_Destroy", None, [ctypes.c_void_p])
    _bind(lib, "Voxels_BoolAdd", None, [ctypes.c_void_p, ctypes.c_void_p])
    _bind(lib, "Voxels_BoolSubtract", None, [ctypes.c_void_p, ctypes.c_void_p])
    _bind(lib, "Voxels_BoolIntersect", None, [ctypes.c_void_p, ctypes.c_void_p])
    _bind_if_exists(lib, "Voxels_BoolAddSmooth", None, [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_float])
    _bind(lib, "Voxels_Offset", None, [ctypes.c_void_p, ctypes.c_float])
    _bind(lib, "Voxels_DoubleOffset", None, [ctypes.c_void_p, ctypes.c_float, ctypes.c_float])
    _bind(lib, "Voxels_TripleOffset", None, [ctypes.c_void_p, ctypes.c_float])
    _bind(lib, "Voxels_Gaussian", None, [ctypes.c_void_p, ctypes.c_float])
    _bind(lib, "Voxels_Median", None, [ctypes.c_void_p, ctypes.c_float])
    _bind(lib, "Voxels_Mean", None, [ctypes.c_void_p, ctypes.c_float])
    _bind(lib, "Voxels_RenderMesh", None, [ctypes.c_void_p, ctypes.c_void_p])
    _bind(lib, "Voxels_RenderImplicit", None, [ctypes.c_void_p, ctypes.POINTER(BBox3), CallbackImplicitDistance])
    _bind(lib, "Voxels_IntersectImplicit", None, [ctypes.c_void_p, CallbackImplicitDistance])
    _bind(lib, "Voxels_RenderLattice", None, [ctypes.c_void_p, ctypes.c_void_p])
    _bind(lib, "Voxels_ProjectZSlice", None, [ctypes.c_void_p, ctypes.c_float, ctypes.c_float])
    _bind_if_exists(lib, "Voxels_bIsInside", ctypes.c_bool, [ctypes.c_void_p, ctypes.POINTER(Vec3)])
    _bind_if_exists(lib, "Voxels_bIsEqual", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_void_p])
    _bind_if_exists(lib, "Voxels_CalculateProperties", None, [ctypes.c_void_p, ctypes.POINTER(ctypes.c_float), ctypes.POINTER(BBox3)])
    _bind_if_exists(lib, "Voxels_GetSurfaceNormal", None, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.POINTER(Vec3)])
    _bind_if_exists(lib, "Voxels_bClosestPointOnSurface", ctypes.c_bool, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.POINTER(Vec3)])
    _bind_if_exists(lib, "Voxels_bRayCastToSurface", ctypes.c_bool, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.POINTER(Vec3), ctypes.POINTER(Vec3)])
    _bind(lib, "Voxels_GetVoxelDimensions", None, [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int)])
    _bind(lib, "Voxels_GetSlice", None, [ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p, ctypes.POINTER(ctypes.c_float)])
    _bind_if_exists(lib, "Voxels_GetInterpolatedSlice", None, [ctypes.c_void_p, ctypes.c_float, ctypes.c_void_p, ctypes.POINTER(ctypes.c_float)])

    # PolyLine
    _bind(lib, "PolyLine_hCreate", ctypes.c_void_p, [ctypes.POINTER(ColorFloat)])
    _bind(lib, "PolyLine_bIsValid", ctypes.c_bool, [ctypes.c_void_p])
    _bind(lib, "PolyLine_Destroy", None, [ctypes.c_void_p])
    _bind(lib, "PolyLine_nAddVertex", ctypes.c_int, [ctypes.c_void_p, ctypes.POINTER(Vec3)])
    _bind(lib, "PolyLine_nVertexCount", ctypes.c_int, [ctypes.c_void_p])
    _bind(lib, "PolyLine_GetVertex", None, [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(Vec3)])
    _bind(lib, "PolyLine_GetColor", None, [ctypes.c_void_p, ctypes.POINTER(ColorFloat)])

    # Viewer bindings kept for API completeness (not used by python viewer).
    _bind_if_exists(lib, "Viewer_hCreate", ctypes.c_void_p, [ctypes.c_char_p, ctypes.POINTER(Vec2), InfoCallback, UpdateCallback, KeyPressedCallback, MouseMovedCallback, MouseButtonCallback, ScrollWheelCallback, WindowSizeCallback])
    _bind_if_exists(lib, "Viewer_bIsValid", ctypes.c_bool, [ctypes.c_void_p])
    _bind_if_exists(lib, "Viewer_Destroy", None, [ctypes.c_void_p])
    _bind_if_exists(lib, "Viewer_RequestUpdate", None, [ctypes.c_void_p])
    _bind_if_exists(lib, "Viewer_bPoll", ctypes.c_bool, [ctypes.c_void_p])
    _bind_if_exists(lib, "Viewer_RequestScreenShot", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p])
    _bind_if_exists(lib, "Viewer_bLoadLightSetup", ctypes.c_bool, [ctypes.c_void_p, ctypes.POINTER(ctypes.c_ubyte), ctypes.c_int, ctypes.POINTER(ctypes.c_ubyte), ctypes.c_int])
    _bind_if_exists(lib, "Viewer_RequestClose", None, [ctypes.c_void_p])
    _bind_if_exists(lib, "Viewer_AddMesh", None, [ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p])
    _bind_if_exists(lib, "Viewer_RemoveMesh", None, [ctypes.c_void_p, ctypes.c_void_p])
    _bind_if_exists(lib, "Viewer_AddPolyLine", None, [ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p])
    _bind_if_exists(lib, "Viewer_RemovePolyLine", None, [ctypes.c_void_p, ctypes.c_void_p])
    _bind_if_exists(lib, "Viewer_SetGroupVisible", None, [ctypes.c_void_p, ctypes.c_int, ctypes.c_bool])
    _bind_if_exists(lib, "Viewer_SetGroupStatic", None, [ctypes.c_void_p, ctypes.c_int, ctypes.c_bool])
    _bind_if_exists(lib, "Viewer_SetGroupMaterial", None, [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(ColorFloat), ctypes.c_float, ctypes.c_float])
    _bind_if_exists(lib, "Viewer_SetGroupMatrix", None, [ctypes.c_void_p, ctypes.c_int, ctypes.POINTER(Mat4)])

    # VDB
    _bind(lib, "VdbFile_hCreate", ctypes.c_void_p, [])
    _bind(lib, "VdbFile_hCreateFromFile", ctypes.c_void_p, [ctypes.c_char_p])
    _bind(lib, "VdbFile_bIsValid", ctypes.c_bool, [ctypes.c_void_p])
    _bind(lib, "VdbFile_Destroy", None, [ctypes.c_void_p])
    _bind(lib, "VdbFile_bSaveToFile", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p])
    _bind(lib, "VdbFile_hGetVoxels", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_int])
    _bind(lib, "VdbFile_nAddVoxels", ctypes.c_int, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p])
    _bind(lib, "VdbFile_hGetScalarField", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_int])
    _bind(lib, "VdbFile_nAddScalarField", ctypes.c_int, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p])
    _bind(lib, "VdbFile_hGetVectorField", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_int])
    _bind(lib, "VdbFile_nAddVectorField", ctypes.c_int, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_void_p])
    _bind(lib, "VdbFile_nFieldCount", ctypes.c_int, [ctypes.c_void_p])
    _bind(lib, "VdbFile_GetFieldName", None, [ctypes.c_void_p, ctypes.c_int, ctypes.c_char_p])
    _bind(lib, "VdbFile_nFieldType", ctypes.c_int, [ctypes.c_void_p, ctypes.c_int])

    # ScalarField
    _bind(lib, "ScalarField_hCreate", ctypes.c_void_p, [])
    _bind(lib, "ScalarField_hCreateCopy", ctypes.c_void_p, [ctypes.c_void_p])
    _bind(lib, "ScalarField_hCreateFromVoxels", ctypes.c_void_p, [ctypes.c_void_p])
    _bind(lib, "ScalarField_hBuildFromVoxels", ctypes.c_void_p, [ctypes.c_void_p, ctypes.c_float, ctypes.c_float])
    _bind(lib, "ScalarField_bIsValid", ctypes.c_bool, [ctypes.c_void_p])
    _bind(lib, "ScalarField_Destroy", None, [ctypes.c_void_p])
    _bind(lib, "ScalarField_SetValue", None, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.c_float])
    _bind(lib, "ScalarField_bGetValue", ctypes.c_bool, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.POINTER(ctypes.c_float)])
    _bind(lib, "ScalarField_RemoveValue", ctypes.c_bool, [ctypes.c_void_p, ctypes.POINTER(Vec3)])
    _bind(lib, "ScalarField_GetVoxelDimensions", None, [ctypes.c_void_p, ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_int)])
    _bind(lib, "ScalarField_GetSlice", None, [ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p])
    _bind(lib, "ScalarField_TraverseActive", None, [ctypes.c_void_p, CallbackScalarFieldTraverse])

    # VectorField
    _bind(lib, "VectorField_hCreate", ctypes.c_void_p, [])
    _bind(lib, "VectorField_hCreateCopy", ctypes.c_void_p, [ctypes.c_void_p])
    _bind(lib, "VectorField_hCreateFromVoxels", ctypes.c_void_p, [ctypes.c_void_p])
    _bind(lib, "VectorField_hBuildFromVoxels", ctypes.c_void_p, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.c_float])
    _bind(lib, "VectorField_bIsValid", ctypes.c_bool, [ctypes.c_void_p])
    _bind(lib, "VectorField_Destroy", None, [ctypes.c_void_p])
    _bind(lib, "VectorField_SetValue", None, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.POINTER(Vec3)])
    _bind(lib, "VectorField_bGetValue", ctypes.c_bool, [ctypes.c_void_p, ctypes.POINTER(Vec3), ctypes.POINTER(Vec3)])
    _bind(lib, "VectorField_RemoveValue", ctypes.c_bool, [ctypes.c_void_p, ctypes.POINTER(Vec3)])
    _bind(lib, "VectorField_TraverseActive", None, [ctypes.c_void_p, CallbackVectorFieldTraverse])

    # Metadata
    _bind(lib, "Metadata_hFromVoxels", ctypes.c_void_p, [ctypes.c_void_p])
    _bind(lib, "Metadata_hFromScalarField", ctypes.c_void_p, [ctypes.c_void_p])
    _bind(lib, "Metadata_hFromVectorField", ctypes.c_void_p, [ctypes.c_void_p])
    _bind(lib, "Metadata_Destroy", None, [ctypes.c_void_p])
    _bind(lib, "Metadata_nCount", ctypes.c_int, [ctypes.c_void_p])
    _bind(lib, "Metadata_nNameLengthAt", ctypes.c_int, [ctypes.c_void_p, ctypes.c_int])
    _bind(lib, "Metadata_bGetNameAt", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_int, ctypes.c_char_p, ctypes.c_int])
    _bind(lib, "Metadata_nTypeAt", ctypes.c_int, [ctypes.c_void_p, ctypes.c_char_p])
    _bind(lib, "Metadata_nStringLengthAt", ctypes.c_int, [ctypes.c_void_p, ctypes.c_char_p])
    _bind(lib, "Metadata_bGetStringAt", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p, ctypes.c_int])
    _bind(lib, "Metadata_bGetFloatAt", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p, ctypes.POINTER(ctypes.c_float)])
    _bind(lib, "Metadata_bGetVectorAt", ctypes.c_bool, [ctypes.c_void_p, ctypes.c_char_p, ctypes.POINTER(Vec3)])
    _bind(lib, "Metadata_SetStringValue", None, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_char_p])
    _bind(lib, "Metadata_SetFloatValue", None, [ctypes.c_void_p, ctypes.c_char_p, ctypes.c_float])
    _bind(lib, "Metadata_SetVectorValue", None, [ctypes.c_void_p, ctypes.c_char_p, ctypes.POINTER(Vec3)])
    _bind(lib, "MetaData_RemoveValue", None, [ctypes.c_void_p, ctypes.c_char_p])


_NATIVE: Native | None = None


def native() -> Native:
    global _NATIVE
    if _NATIVE is None:
        lib = _load_lib()
        _configure(lib)
        _NATIVE = Native(lib=lib)
    return _NATIVE



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from _common.types import Vector3Like

if TYPE_CHECKING:
    from .mesh import Mesh


class MeshMath:
    @staticmethod
    def bPointLiesOnTriangle(vecP: Vector3Like, vecA: Vector3Like, vecB: Vector3Like, vecC: Vector3Like) -> bool:
        p = np.array(vecP, dtype=np.float64)
        a = np.array(vecA, dtype=np.float64) - p
        b = np.array(vecB, dtype=np.float64) - p
        c = np.array(vecC, dtype=np.float64) - p
        u = np.cross(b, c)
        v = np.cross(c, a)
        w = np.cross(a, b)
        return float(np.dot(u, v)) >= 0.0 and float(np.dot(u, w)) >= 0.0

    @staticmethod
    def bFindTriangleFromSurfacePoint(mesh: Mesh, vecSurfacePoint: Vector3Like) -> tuple[bool, int]:
        for i in range(mesh.nTriangleCount()):
            a, b, c = mesh.GetTriangle(i)
            if MeshMath.bPointLiesOnTriangle(vecSurfacePoint, a, b, c):
                return (True, i)
        return (False, (2**31) - 1)




from __future__ import annotations

import io
import math
import threading
import time
import zipfile
from pathlib import Path
from queue import Empty, Queue
from typing import Any, Optional, Sequence, cast

import numpy as np
import vedo
from vedo import Mesh as VedoMesh

from .._types import ColorFloat
from .._core import Mesh, PolyLine, Voxels
from .keybindings import IKeyHandler, KeyAction, KeyHandler, RotateToNextRoundAngleAction
from .timelapse import _TimeLapseState

class VedoViewer:
    """C#-style viewer facade adapted to vedo with a thread-safe command queue."""

    def __init__(self, title: str = "PicoGK", offscreen: bool = False) -> None:
        self._plotter = vedo.Plotter(title=title, interactive=False, offscreen=offscreen)
        self._queue: Queue[tuple[str, object]] = Queue()
        self._running = False
        self._closed = False
        self._thread: Optional[threading.Thread] = None
        self._actors: dict[int, object] = {}
        self._groups: dict[int, set[int]] = {}
        self._object_to_actor: dict[int, int] = {}
        self._group_visible: dict[int, bool] = {}
        self._group_static: dict[int, bool] = {}
        self._group_material: dict[int, tuple[tuple[float, float, float, float], float, float]] = {}
        self._group_matrix: dict[int, tuple[float, ...]] = {}
        self._next_actor_id = 1
        self._idle = True
        self._light_setup_raw: tuple[bytes, bytes] | None = None
        self._timelapse: _TimeLapseState | None = None
        self._animations: list[Any] = []
        self._key_handlers: list[IKeyHandler] = []
        self._key_callback_handle: object | None = None

        self.m_fElevation = 30.0
        self.m_fOrbit = 45.0
        self.m_fFov = 45.0
        self.m_fZoom = 1.0
        self.m_bPerspective = True

        self._default_key_handler = KeyHandler()
        self._default_key_handler.AddAction(KeyAction(RotateToNextRoundAngleAction(RotateToNextRoundAngleAction.EDir.Dir_Down), "Down"))
        self._default_key_handler.AddAction(KeyAction(RotateToNextRoundAngleAction(RotateToNextRoundAngleAction.EDir.Dir_Up), "Up"))
        self._default_key_handler.AddAction(KeyAction(RotateToNextRoundAngleAction(RotateToNextRoundAngleAction.EDir.Dir_Left), "Left"))
        self._default_key_handler.AddAction(KeyAction(RotateToNextRoundAngleAction(RotateToNextRoundAngleAction.EDir.Dir_Right), "Right"))
        self.AddKeyHandler(self._default_key_handler)

    def _enqueue(self, cmd: str, payload: object = None) -> None:
        self._idle = False
        self._queue.put((cmd, payload))

    def _stop_loop(self) -> None:
        self._running = False
        if self._thread is not None:
            self._thread.join(timeout=2.0)
            self._thread = None

    def start(self) -> None:
        if self._running:
            return
        self._running = True
        self._thread = threading.Thread(target=self._loop, daemon=True)
        self._thread.start()

    def _is_visible(self, actor: object) -> bool:
        visible_fn = getattr(actor, "is_visible", None)
        if callable(visible_fn):
            try:
                return bool(visible_fn())
            except Exception:
                return True
        return True

    def _set_actor_visible(self, actor: object, visible: bool) -> None:
        if visible:
            on_fn = getattr(actor, "on", None)
            if callable(on_fn):
                on_fn()
                return
        else:
            off_fn = getattr(actor, "off", None)
            if callable(off_fn):
                off_fn()
                return

        set_vis = getattr(actor, "SetVisibility", None)
        if callable(set_vis):
            set_vis(bool(visible))

    def _set_actor_pickable(self, actor: object, pickable: bool) -> None:
        pickable_fn = getattr(actor, "pickable", None)
        if callable(pickable_fn):
            try:
                pickable_fn(bool(pickable))
                return
            except Exception:
                pass
        set_pickable = getattr(actor, "SetPickable", None)
        if callable(set_pickable):
            set_pickable(bool(pickable))

    def _apply_matrix_points(self, actor: object, mat: tuple[float, ...]) -> None:
        if len(mat) != 16:
            raise ValueError("mat must contain 16 values")
        pts_fn = getattr(actor, "points", None)
        if not callable(pts_fn):
            return
        try:
            points = np.asarray(pts_fn(), dtype=np.float64)
        except Exception:
            return
        if points.ndim != 2 or points.shape[0] == 0 or points.shape[1] != 3:
            return
        m = np.array(mat, dtype=np.float64).reshape((4, 4))
        homo = np.hstack([points, np.ones((points.shape[0], 1), dtype=np.float64)])
        transformed = (homo @ m.T)[:, :3]
        pts_fn(transformed.astype(np.float32))

    def _apply_group_style_to_actor(self, actor_id: int) -> None:
        actor = self._actors.get(actor_id)
        if actor is None:
            return
        group_id = None
        for gid, members in self._groups.items():
            if actor_id in members:
                group_id = gid
                break
        if group_id is None:
            return

        visible = self._group_visible.get(group_id)
        if visible is not None:
            self._set_actor_visible(actor, visible)

        static = self._group_static.get(group_id)
        if static is not None:
            self._set_actor_pickable(actor, not static)

        material = self._group_material.get(group_id)
        if material is not None:
            clr, _metallic, _roughness = material
            color_fn = getattr(actor, "c", None)
            if callable(color_fn):
                color_fn((clr[0], clr[1], clr[2]))
            alpha_fn = getattr(actor, "alpha", None)
            if callable(alpha_fn):
                alpha_fn(clr[3])

        matrix = self._group_matrix.get(group_id)
        if matrix is not None:
            self._apply_matrix_points(actor, matrix)

    def _active_bounds(self) -> tuple[float, float, float, float, float, float] | None:
        items: list[tuple[float, float, float, float, float, float]] = []
        for actor in self._actors.values():
            if not self._is_visible(actor):
                continue
            bounds_fn = getattr(actor, "bounds", None)
            if not callable(bounds_fn):
                continue
            try:
                raw_bounds = cast(Sequence[object], bounds_fn())
                b = tuple(float(cast(Any, v)) for v in raw_bounds)
            except Exception:
                continue
            if len(b) == 6:
                items.append(cast(tuple[float, float, float, float, float, float], b))
        if not items:
            return None
        xmin = min(v[0] for v in items)
        xmax = max(v[1] for v in items)
        ymin = min(v[2] for v in items)
        ymax = max(v[3] for v in items)
        zmin = min(v[4] for v in items)
        zmax = max(v[5] for v in items)
        return (xmin, xmax, ymin, ymax, zmin, zmax)

    def _update_camera(self) -> None:
        bounds = self._active_bounds()
        if bounds is None:
            return
        xmin, xmax, ymin, ymax, zmin, zmax = bounds
        cx = (xmin + xmax) * 0.5
        cy = (ymin + ymax) * 0.5
        cz = (zmin + zmax) * 0.5
        rx = max(1e-3, xmax - xmin)
        ry = max(1e-3, ymax - ymin)
        rz = max(1e-3, zmax - zmin)
        radius = max(rx, ry, rz) * 1.5 * max(1e-3, self.m_fZoom)

        orbit = math.radians(self.m_fOrbit)
        elev = math.radians(self.m_fElevation)
        cos_e = math.cos(elev)
        ex = cx + (math.cos(orbit) * cos_e * radius)
        ey = cy + (math.sin(orbit) * cos_e * radius)
        ez = cz + (math.sin(elev) * radius)

        cam = self._plotter.camera
        cam.SetPosition(ex, ey, ez)
        cam.SetFocalPoint(cx, cy, cz)
        cam.SetViewUp(0.0, 0.0, 1.0)
        cam.SetViewAngle(float(self.m_fFov))
        if self.m_bPerspective:
            cam.ParallelProjectionOff()
        else:
            cam.ParallelProjectionOn()

    def _pulse_animations(self) -> bool:
        if not self._animations:
            return False
        current = time.perf_counter() - self._anim_start
        keep: list[Any] = []
        changed = False
        for anim in self._animations:
            try:
                alive = bool(anim.bAnimate(float(current)))
            except Exception:
                alive = False
            changed = True
            if alive:
                keep.append(anim)
        self._animations = keep
        return changed

    def _capture_timelapse_if_due(self) -> bool:
        if self._timelapse is None:
            return False
        due, path = self._timelapse.bDue()
        if not due:
            return False
        self._plotter.screenshot(path)
        return True

    def _render(self) -> None:
        self._update_camera()
        self._plotter.render(resetcam=False)

    def _next_id(self) -> int:
        out = self._next_actor_id
        self._next_actor_id += 1
        return out

    def _process_command(self, cmd: str, payload: object) -> None:
        if cmd == "close":
            self._running = False
            self._closed = True
            return

        if cmd == "add_actor":
            actor_id, actor, group_id, object_key = cast(tuple[int, object, int, int], payload)
            self._actors[actor_id] = self._plotter.add(actor)
            self._groups.setdefault(group_id, set()).add(actor_id)
            if object_key >= 0:
                self._object_to_actor[object_key] = actor_id
            self._apply_group_style_to_actor(actor_id)
            self._render()
            return

        if cmd == "remove_actor":
            actor_id = int(cast(int, payload))
            actor = self._actors.pop(actor_id, None)
            if actor is not None:
                self._plotter.remove(actor)
            for members in self._groups.values():
                members.discard(actor_id)
            keys = [k for k, v in self._object_to_actor.items() if v == actor_id]
            for k in keys:
                self._object_to_actor.pop(k, None)
            self._render()
            return

        if cmd == "remove_object":
            object_key = int(cast(int, payload))
            actor_id = self._object_to_actor.pop(object_key, None)
            if actor_id is not None:
                self._process_command("remove_actor", actor_id)
            return

        if cmd == "remove_all":
            for actor in list(self._actors.values()):
                self._plotter.remove(actor)
            self._actors.clear()
            self._groups.clear()
            self._object_to_actor.clear()
            self._render()
            return

        if cmd == "render":
            self._render()
            return

        if cmd == "set_group_visible":
            group_id, visible = cast(tuple[int, bool], payload)
            self._group_visible[group_id] = bool(visible)
            for actor_id in self._groups.get(group_id, set()):
                actor = self._actors.get(actor_id)
                if actor is not None:
                    self._set_actor_visible(actor, visible)
            self._render()
            return

        if cmd == "set_group_static":
            group_id, static = cast(tuple[int, bool], payload)
            self._group_static[group_id] = bool(static)
            for actor_id in self._groups.get(group_id, set()):
                actor = self._actors.get(actor_id)
                if actor is not None:
                    self._set_actor_pickable(actor, not static)
            return

        if cmd == "set_group_material":
            group_id, clr, metallic, roughness = cast(tuple[int, tuple[float, float, float, float], float, float], payload)
            self._group_material[group_id] = (clr, metallic, roughness)
            for actor_id in self._groups.get(group_id, set()):
                self._apply_group_style_to_actor(actor_id)
            self._render()
            return

        if cmd == "set_group_matrix":
            group_id, mat = cast(tuple[int, tuple[float, ...]], payload)
            self._group_matrix[group_id] = mat
            for actor_id in self._groups.get(group_id, set()):
                self._apply_group_style_to_actor(actor_id)
            self._render()
            return

        if cmd == "set_background":
            r, g, b = cast(tuple[float, float, float], payload)
            self._plotter.background((r, g, b))
            self._render()
            return

        if cmd == "set_view":
            orbit, elevation = cast(tuple[float, float], payload)
            self.m_fOrbit = float(orbit)
            self.m_fElevation = float(max(-89.9, min(89.9, elevation)))
            self._render()
            return

        if cmd == "set_fov":
            self.m_fFov = float(cast(float, payload))
            self._render()
            return

        if cmd == "set_zoom":
            self.m_fZoom = float(max(0.05, cast(float, payload)))
            self._render()
            return

        if cmd == "set_perspective":
            self.m_bPerspective = bool(cast(bool, payload))
            self._render()
            return

        if cmd == "screenshot":
            path = str(cast(str, payload))
            Path(path).parent.mkdir(parents=True, exist_ok=True)
            self._plotter.screenshot(path)
            return

        if cmd == "log_stats":
            bounds = self._active_bounds()
            print(f"Viewer stats: actors={len(self._actors)} groups={len(self._groups)} queue={self._queue.qsize()} idle={self._idle}")
            if bounds is not None:
                print(f"Bounds: x[{bounds[0]:.3f},{bounds[1]:.3f}] y[{bounds[2]:.3f},{bounds[3]:.3f}] z[{bounds[4]:.3f},{bounds[5]:.3f}]")
            return

        if cmd == "light_setup":
            self._light_setup_raw = cast(tuple[bytes, bytes], payload)
            return

    def _drain_queue(self) -> None:
        drained = False
        while True:
            try:
                cmd, payload = self._queue.get_nowait()
            except Empty:
                break
            drained = True
            self._process_command(cmd, payload)
        if drained:
            self._idle = self._queue.empty()

    def _loop(self) -> None:
        while self._running:
            updated = False
            try:
                cmd, payload = self._queue.get(timeout=0.03)
                self._process_command(cmd, payload)
                updated = True
            except Empty:
                pass
            if self._pulse_animations():
                updated = True
            if self._capture_timelapse_if_due():
                updated = True
            if updated:
                self._render()
            self._idle = self._queue.empty()

    def _color_tuple(self, clr: ColorFloat | Sequence[float]) -> tuple[float, float, float, float]:
        if isinstance(clr, ColorFloat):
            return (float(clr.r), float(clr.g), float(clr.b), float(clr.a))
        if len(clr) == 3:
            return (float(clr[0]), float(clr[1]), float(clr[2]), 1.0)
        if len(clr) == 4:
            return (float(clr[0]), float(clr[1]), float(clr[2]), float(clr[3]))
        raise ValueError("Color must be (r,g,b) or (r,g,b,a)")

    def _make_mesh_actor(self, mesh: Mesh, color: str | Sequence[float] = "lightgray") -> object:
        vertices, triangles = mesh.to_numpy()
        # vedo expects plain triangle connectivity [v0, v1, v2] per face.
        # Passing VTK-style [3, v0, v1, v2] corrupts topology in rendering.
        faces = np.ascontiguousarray(triangles, dtype=np.int32)
        actor = VedoMesh([vertices, faces])
        # Do not force a normals recomputation here: runtime meshes may contain
        # local winding inconsistencies, and averaging normals can produce
        # severe shading artifacts (striping / cone-like appearance).
        phong_fn = getattr(actor, "phong", None)
        if callable(phong_fn):
            phong_fn()
        if isinstance(color, str):
            actor.c(color)
        else:
            clr = self._color_tuple(cast(Sequence[float], color))
            actor.c((clr[0], clr[1], clr[2])).alpha(clr[3])
        return actor

    def _make_polyline_actor(self, poly: PolyLine) -> object:
        n = poly.vertex_count()
        color = poly.get_color()
        if n <= 0:
            actor = vedo.Points(np.array([[0.0, 0.0, 0.0]], dtype=np.float32), c=(color[0], color[1], color[2]), r=1)
            alpha_fn = getattr(actor, "alpha", None)
            if callable(alpha_fn):
                alpha_fn(color[3])
            return actor
        pts = np.array([poly.get_vertex(i) for i in range(n)], dtype=np.float32)
        if n == 1:
            actor = vedo.Points(pts, c=(color[0], color[1], color[2]), r=8)
            alpha_fn = getattr(actor, "alpha", None)
            if callable(alpha_fn):
                alpha_fn(color[3])
            return actor
        actor = vedo.Line(pts, lw=2)
        color_fn = getattr(actor, "c", None)
        if callable(color_fn):
            color_fn((color[0], color[1], color[2]))
        alpha_fn = getattr(actor, "alpha", None)
        if callable(alpha_fn):
            alpha_fn(color[3])
        return actor

    def Add(self, xObject: object, nGroupID: int = 0) -> int:
        if isinstance(xObject, Mesh):
            return self.add_mesh(xObject, nGroupID=nGroupID)
        if isinstance(xObject, Voxels):
            return self.add_voxels(xObject, nGroupID=nGroupID)
        if isinstance(xObject, PolyLine):
            return self.add_polyline(xObject, nGroupID=nGroupID)
        raise TypeError("Add expects Mesh, Voxels, or PolyLine")

    def add_mesh(self, mesh: Mesh, color: str | Sequence[float] = "lightgray", nGroupID: int = 0) -> int:
        actor_id = self._next_id()
        actor = self._make_mesh_actor(mesh, color)
        self._enqueue("add_actor", (actor_id, actor, int(nGroupID), id(mesh)))
        return actor_id

    def add_voxels(self, vox: Voxels, nGroupID: int = 0, color: str | Sequence[float] = "lightgray") -> int:
        with Mesh.from_voxels(vox) as mesh:
            actor = self._make_mesh_actor(mesh, color)
        actor_id = self._next_id()
        self._enqueue("add_actor", (actor_id, actor, int(nGroupID), id(vox)))
        return actor_id

    def add_polyline(self, poly: PolyLine, nGroupID: int = 0) -> int:
        actor_id = self._next_id()
        actor = self._make_polyline_actor(poly)
        self._enqueue("add_actor", (actor_id, actor, int(nGroupID), id(poly)))
        return actor_id

    def Remove(self, xObject: object) -> None:
        if isinstance(xObject, int):
            self._enqueue("remove_actor", int(xObject))
            return
        self._enqueue("remove_object", id(xObject))

    def remove(self, actor_id: int) -> None:
        self.Remove(actor_id)

    def RemoveAllObjects(self) -> None:
        self._enqueue("remove_all")

    def RequestUpdate(self) -> None:
        self._enqueue("render")

    def request_render(self) -> None:
        self.RequestUpdate()

    def RequestScreenShot(self, strScreenShotPath: str) -> None:
        self._enqueue("screenshot", str(strScreenShotPath))

    def SetGroupVisible(self, nGroupID: int, bVisible: bool) -> None:
        self._enqueue("set_group_visible", (int(nGroupID), bool(bVisible)))

    def SetGroupStatic(self, nGroupID: int, bStatic: bool) -> None:
        self._enqueue("set_group_static", (int(nGroupID), bool(bStatic)))

    def SetGroupMaterial(self, nGroupID: int, clr: ColorFloat | Sequence[float], fMetallic: float, fRoughness: float) -> None:
        self._enqueue("set_group_material", (int(nGroupID), self._color_tuple(clr), float(fMetallic), float(fRoughness)))

    def SetGroupMatrix(self, nGroupID: int, mat: Sequence[float]) -> None:
        if len(mat) != 16:
            raise ValueError("mat must contain 16 values (row-major 4x4)")
        self._enqueue("set_group_matrix", (int(nGroupID), tuple(float(v) for v in mat)))

    def SetBackgroundColor(self, clr: ColorFloat | Sequence[float]) -> None:
        rgba = self._color_tuple(clr)
        self._enqueue("set_background", (rgba[0], rgba[1], rgba[2]))

    def AdjustViewAngles(self, fOrbitRelative: float, fElevationRelative: float) -> None:
        self.SetViewAngles(self.m_fOrbit + float(fOrbitRelative), self.m_fElevation + float(fElevationRelative))

    def SetViewAngles(self, fOrbit: float, fElevation: float) -> None:
        orbit = float(fOrbit)
        while orbit >= 360.0:
            orbit -= 360.0
        while orbit < 0.0:
            orbit += 360.0
        elev = max(-89.9, min(89.9, float(fElevation)))
        self._enqueue("set_view", (orbit, elev))

    def SetFov(self, fAngle: float) -> None:
        self._enqueue("set_fov", float(fAngle))

    def SetZoom(self, fZoom: float) -> None:
        self._enqueue("set_zoom", float(fZoom))

    def SetPerspective(self, bPerspective: bool) -> None:
        self._enqueue("set_perspective", bool(bPerspective))

    def LogStatistics(self) -> None:
        self._enqueue("log_stats")

    def LoadLightSetup(self, source: str | bytes | io.BytesIO) -> None:
        if isinstance(source, str):
            blob = Path(source).read_bytes()
        elif isinstance(source, io.BytesIO):
            blob = source.getvalue()
        else:
            blob = bytes(source)

        with zipfile.ZipFile(io.BytesIO(blob), "r") as zf:
            diffuse = zf.read("Diffuse.dds")
            specular = zf.read("Specular.dds")
        self._enqueue("light_setup", (diffuse, specular))

    def AddKeyHandler(self, xKeyHandler: IKeyHandler) -> None:
        self._key_handlers.insert(0, xKeyHandler)

    def _ensure_key_callback(self) -> None:
        if self._key_callback_handle is not None:
            return

        def _on_key(event: object) -> None:
            key = str(getattr(event, "keypress", getattr(event, "keyPressed", "")))
            if not key:
                return
            shift = bool(getattr(event, "shift", False))
            ctrl = bool(getattr(event, "ctrl", False))
            alt = bool(getattr(event, "alt", False))
            cmd = bool(getattr(event, "cmd", False))
            for handler in self._key_handlers:
                if handler.bHandleEvent(self, key, False, shift, ctrl, alt, cmd):
                    self.RequestUpdate()
                    break

        self._key_callback_handle = self._plotter.add_callback("KeyPress", _on_key)

    def AddAnimation(self, oAnim: Any) -> None:
        if not hasattr(self, "_anim_start"):
            self._anim_start = time.perf_counter()
        self._animations.append(oAnim)
        self.RequestUpdate()

    def RemoveAllAnimations(self) -> None:
        for anim in self._animations:
            end_fn = getattr(anim, "End", None)
            if callable(end_fn):
                end_fn()
        self._animations.clear()

    def StartTimeLapse(
        self,
        fIntervalInMilliseconds: float,
        strPath: str,
        strFileName: str = "frame_",
        nStartFrame: int = 0,
        bPaused: bool = False,
    ) -> None:
        self._timelapse = _TimeLapseState.create(
            fIntervalInMilliseconds,
            strPath,
            strFileName,
            nStartFrame,
            bPaused,
        )

    def PauseTimeLapse(self) -> None:
        if self._timelapse is not None:
            self._timelapse.Pause()

    def ResumeTimeLapse(self) -> None:
        if self._timelapse is not None:
            self._timelapse.Resume()

    def StopTimeLapse(self) -> None:
        self._timelapse = None

    def bPoll(self) -> bool:
        self._drain_queue()
        changed = self._pulse_animations()
        changed = self._capture_timelapse_if_due() or changed
        if changed:
            self._render()
        self._idle = self._queue.empty()
        return not self._closed

    def bIsIdle(self) -> bool:
        self._idle = self._idle and self._queue.empty()
        return self._idle

    def interact(self) -> None:
        self._stop_loop()
        self._drain_queue()
        self._ensure_key_callback()
        # Phase 1: open the VTK window non-interactively so the render window
        # and camera object actually exist.  resetcam=True lets vedo fit the
        # scene so the object is always visible before we reposition.
        self._plotter.show(interactive=False, resetcam=True)
        # Phase 2: now that the window is live, apply our computed camera
        # position (respecting m_fZoom, m_fOrbit, m_fElevation, m_fFov) on
        # the real VTK camera, then re-render without resetting.
        self._update_camera()
        self._plotter.render(resetcam=False)
        # Phase 3: hand off to the VTK interactive event loop.
        if getattr(self._plotter, "interactor", None) is not None:
            self._plotter.interactor.Start()
        else:
            self._plotter.show(interactive=True, resetcam=False)

    def close(self) -> None:
        if self._closed:
            return
        self._enqueue("close")
        self._stop_loop()
        self._drain_queue()
        self._plotter.close()






from __future__ import annotations

from collections.abc import Callable
from enum import Enum

from picogk import Image
from .line_modulation import LineModulation


class SurfaceModulation:
    class EInput(Enum):
        FUNC = "FUNC"
        IMAGE = "IMAGE"

    class ELine(Enum):
        FIRST = "FIRST"
        SECOND = "SECOND"

    MappingFunc = Callable[[float], float]
    RatioFunc = Callable[[float, float], float]

    def __init__(self, source: float | RatioFunc) -> None:
        self.m_fConstValue = 0.0
        self.m_oFunc: SurfaceModulation.RatioFunc
        self.m_eInput = self.EInput.FUNC
        self.m_eLine = self.ELine.SECOND
        self.m_oLineModulation: LineModulation | None = None
        self.m_oImage: Image | None = None
        self.m_oMappingFunc: SurfaceModulation.MappingFunc | None = None

        if callable(source):
            self.m_oFunc = source
            return

        self.m_fConstValue = float(source)
        self.m_oFunc = self._const_surface_func

    @classmethod
    def from_line_modulation(
        cls,
        oLineModulation: LineModulation,
        eLine: "SurfaceModulation.ELine" = ELine.SECOND,
    ) -> "SurfaceModulation":
        mod = cls(0.0)
        mod.m_oLineModulation = oLineModulation
        mod.m_eLine = eLine
        mod.m_oFunc = mod._line_surface_func
        mod.m_eInput = mod.EInput.FUNC
        return mod

    @classmethod
    def from_image(cls, oImage: Image, oMappingFunc: MappingFunc) -> "SurfaceModulation":
        mod = cls(0.0)
        mod.m_oImage = oImage
        mod.m_oMappingFunc = oMappingFunc
        mod.m_eInput = mod.EInput.IMAGE
        return mod

    def _const_surface_func(self, fPhi: float, fLengthRatio: float) -> float:
        return self.m_fConstValue

    def _line_surface_func(self, fPhi: float, fLengthRatio: float) -> float:
        if self.m_oLineModulation is None:
            raise RuntimeError("Line modulation source is not configured")
        if self.m_eLine == self.ELine.FIRST:
            return self.m_oLineModulation.fGetModulation(fPhi)
        return self.m_oLineModulation.fGetModulation(fLengthRatio)

    def oGetInputType(self) -> "SurfaceModulation.EInput":
        return self.m_eInput

    def fGetModulation(self, fPhi: float, fLengthRatio: float) -> float:
        if self.m_eInput == self.EInput.FUNC:
            return float(self.m_oFunc(float(fPhi), float(fLengthRatio)))

        if self.m_oImage is None or self.m_oMappingFunc is None:
            raise RuntimeError("Image modulation source is not configured")

        nXRange = int(self.m_oImage.nWidth) - 1
        nYRange = int(self.m_oImage.nHeight) - 1
        x = int(round(nXRange - fPhi * nXRange))
        y = int(round(fLengthRatio * nYRange))
        gray = float(self.m_oImage.fValue(x, y))
        return float(self.m_oMappingFunc(gray))

    def f_get_modulation(self, fPhi: float, fLengthRatio: float) -> float:
        return self.fGetModulation(fPhi, fLengthRatio)

    def __mul__(self, fFactor: float) -> "SurfaceModulation":
        return SurfaceModulation(lambda phi, lr: float(fFactor) * self.fGetModulation(phi, lr))

    def __rmul__(self, fFactor: float) -> "SurfaceModulation":
        return self.__mul__(fFactor)

    def __add__(self, other: "SurfaceModulation") -> "SurfaceModulation":
        return SurfaceModulation(lambda phi, lr: self.fGetModulation(phi, lr) + other.fGetModulation(phi, lr))

    def __sub__(self, other: "SurfaceModulation") -> "SurfaceModulation":
        return SurfaceModulation(lambda phi, lr: self.fGetModulation(phi, lr) - other.fGetModulation(phi, lr))




import csv
from pathlib import Path
from typing import Protocol, Sequence


class IDataTable(Protocol):
    def nRowCount(self) -> int:
        ...

    def nMaxColumnCount(self) -> int:
        ...

    def strGetAt(self, nRow: int, nCol: int) -> str:
        ...


class CsvTable(IDataTable):
    def __init__(self, source: str | Path | IDataTable | None = None) -> None:
        self._rows: list[list[str]] = []
        self._key_column = -1
        self._col_ids: dict[str, int] = {}
        if source is None:
            return
        if isinstance(source, (str, Path)):
            with open(source, "r", newline="", encoding="utf-8") as handle:
                reader = csv.reader(handle)
                self._rows = [list(row) for row in reader]
        else:
            for r in range(source.nRowCount()):
                row = [source.strGetAt(r, c) for c in range(source.nMaxColumnCount())]
                self._rows.append(row)

    def Save(self, file_path: str | Path) -> None:
        with open(file_path, "w", newline="", encoding="utf-8") as handle:
            writer = csv.writer(handle)
            writer.writerows(self._rows)

    def nRowCount(self) -> int:
        return len(self._rows)

    def nMaxColumnCount(self) -> int:
        return max((len(r) for r in self._rows), default=0)

    def strGetAt(self, nRow: int, nCol: int) -> str:
        if nRow < 0 or nRow >= len(self._rows):
            return ""
        row = self._rows[nRow]
        if nCol < 0 or nCol >= len(row):
            return ""
        return row[nCol]

    def SetKeyColumn(self, nColumn: int) -> None:
        self._key_column = int(nColumn)

    def bGetAt(self, key_or_row: str | int, col_or_idx: str | int) -> tuple[bool, str]:
        if isinstance(key_or_row, int) and isinstance(col_or_idx, int):
            value = self.strGetAt(key_or_row, col_or_idx)
            return (value != "", value)
        if not isinstance(key_or_row, str) or not isinstance(col_or_idx, str):
            return (False, "")
        if self._key_column < 0:
            return (False, "")
        ok_col, col = self.bFindColumn(col_or_idx)
        if not ok_col:
            return (False, "")
        for row in self._rows:
            if self._key_column < len(row) and row[self._key_column] == key_or_row:
                if col < len(row):
                    return (True, row[col])
                return (False, "")
        return (False, "")

    def bFindColumn(self, strColumnID: str) -> tuple[bool, int]:
        idx = self._col_ids.get(strColumnID)
        if idx is None:
            return (False, -1)
        return (True, idx)

    def strColumnId(self, nColumn: int) -> str:
        for key, val in self._col_ids.items():
            if val == nColumn:
                return key
        return ""

    def SetColumnIds(self, ids: Sequence[str]) -> None:
        self._col_ids = {str(col): i for i, col in enumerate(ids)}

    def AddRow(self, values: Sequence[str]) -> None:
        self._rows.append([str(v) for v in values])





from __future__ import annotations

import numpy as np

from picogk import Voxels
from shape_kernel._types import Vector3Like, as_np3
from shape_kernel.utils import Uf

from .i_beam_thickness import IBeamThickness
from ..unit_cells import IUnitCell


class BoundaryBeamThickness(IBeamThickness):
	def __init__(self, fMinBeamThickness: float, fMaxBeamThickness: float) -> None:
		self.m_fMinBeamThickness = float(fMinBeamThickness)
		self.m_fMaxBeamThickness = float(fMaxBeamThickness)
		self.m_voxBounding: Voxels | None = None

	def fGetBeamThickness(self, vecPt: Vector3Like) -> float:
		if self.m_voxBounding is None:
			raise ValueError("No Boundary Voxels specified.")

		bSuccess, vecSurf = self.m_voxBounding.bClosestPointOnSurface(vecPt)
		if bSuccess is False:
			raise ValueError("No Closest Point found.")

		fDistance = float(np.linalg.norm(as_np3(vecSurf) - as_np3(vecPt)))
		return Uf.fTransSmooth(self.m_fMaxBeamThickness, self.m_fMinBeamThickness, fDistance, 15.0, 5.0)

	def UpdateCell(self, xCell: IUnitCell) -> None:
		return None

	def SetBoundingVoxels(self, voxBounding: Voxels) -> None:
		self.m_voxBounding = voxBounding


__all__ = ["BoundaryBeamThickness"]




import datetime
from pathlib import Path
import time


class LogFile:
    def __init__(self, strFileName: str = "", bOutputToConsole: bool = True) -> None:
        self._output_to_console = bool(bOutputToConsole)
        self._start = time.perf_counter()
        self._last = 0.0
        if not strFileName:
            docs = Path.home() / "Documents"
            docs.mkdir(parents=True, exist_ok=True)
            stamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            strFileName = str(docs / f"PicoGK_{stamp}.log")
        self._path = strFileName
        self._writer = open(strFileName, "w", encoding="utf-8")
        self.Log(f"Opened {strFileName}")

    def Log(self, strFormat: str, *args: object) -> None:
        now = time.perf_counter() - self._start
        diff = now - self._last
        msg = strFormat.format(*args) if args else strFormat
        prefix = f"{now:7.0f}s {diff:6.1f}+ "
        for line in msg.split("\n"):
            row = prefix + line
            if self._output_to_console:
                print(row)
            self._writer.write(row + "\n")
        self._writer.flush()
        self._last = now

    def LogTime(self) -> None:
        self.Log("Current time (UTC): {0}", datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S (UTC)"))
        self.Log("Current local time: {0}", datetime.datetime.now().astimezone().strftime("%Y-%m-%d %H:%M:%S (%z)"))

    def dispose(self) -> None:
        if self._writer.closed:
            return
        self.Log("Closing log file.")
        self.LogTime()
        self._writer.close()

    def __enter__(self) -> "LogFile":
        return self

    def __exit__(self, exc_type: object, exc: object, tb: object) -> None:
        self.dispose()



from __future__ import annotations

from typing import Any, Sequence, TypeAlias

import numpy as np
from numpy.typing import NDArray

NumericArray: TypeAlias = NDArray[np.floating[Any]] | NDArray[np.integer[Any]]
FloatSequence: TypeAlias = Sequence[float]

Vector2Like: TypeAlias = FloatSequence | NumericArray
Vector3Like: TypeAlias = FloatSequence | NumericArray
ColorLike: TypeAlias = FloatSequence | NumericArray


def as_np3(value: Vector3Like) -> NDArray[np.float64]:
    arr = np.asarray(value, dtype=np.float64)
    if arr.shape != (3,):
        raise ValueError("Expected a 3D vector")
    return arr


def as_tuple3(value: Vector3Like) -> tuple[float, float, float]:
    arr = as_np3(value)
    return (float(arr[0]), float(arr[1]), float(arr[2]))


def np_vec3(value: Vector3Like) -> NDArray[np.float64]:
    return as_np3(value)


def clamp01(value: float) -> float:
    return float(np.clip(float(value), 0.0, 1.0))


def normalized(value: Vector3Like, eps: float = 1e-12) -> NDArray[np.float64]:
    arr = as_np3(value)
    length = float(np.linalg.norm(arr))
    if length <= eps:
        return np.zeros(3, dtype=np.float64)
    return arr / length


from __future__ import annotations

from picogk import Lattice, Voxels

from .._types import Vec3
from ..frames import Frames
from ..frames.local_frames import LocalFrame
from ..modulations import LineModulation
from .base_shape import BaseShape


class LatticePipe(BaseShape):
	def __init__(
		self,
		aFrames: Frames,
		fRadius: float,
		*,
		_length_steps: int,
	) -> None:
		super().__init__()
		self.m_aFrames = aFrames
		self.SetLengthSteps(_length_steps)
		self.m_oRadiusModulation = LineModulation(fRadius)

	@classmethod
	def from_frame(
		cls,
		oFrame: LocalFrame,
		fLength: float = 20.0,
		fRadius: float = 10.0,
	) -> "LatticePipe":
		return cls(Frames.from_length(fLength, oFrame), fRadius, _length_steps=100)

	@classmethod
	def from_frames(
		cls,
		aFrames: Frames,
		fRadius: float = 10.0,
	) -> "LatticePipe":
		return cls(aFrames, fRadius, _length_steps=500)

	def SetRadius(self, oModulation: LineModulation) -> None:
		self.m_oRadiusModulation = oModulation

	def SetLengthSteps(self, nLengthSteps: int) -> None:
		self.m_nLengthSteps = int(nLengthSteps)

	def voxConstruct(self) -> Voxels:
		return Voxels.from_lattice(self.latConstruct())

	def latConstruct(self) -> Lattice:
		lattice = Lattice()
		for iz in range(1, self.m_nLengthSteps):
			lr0 = (1.0 / self.m_nLengthSteps) * (iz - 1)
			lr1 = (1.0 / self.m_nLengthSteps) * iz
			p0 = self.vecGetSpinePoint(lr0)
			p1 = self.vecGetSpinePoint(lr1)
			b0 = self.fGetRadius(lr0)
			b1 = self.fGetRadius(lr1)
			lattice.add_beam(p0, p1, b0, b1)
		return lattice

	def vecGetSpinePoint(self, fLengthRatio: float) -> Vec3:
		return self.m_fnTrafo(self.m_aFrames.vecGetSpineAlongLength(fLengthRatio))

	def fGetRadius(self, fLengthRatio: float) -> float:
		return self.m_oRadiusModulation.fGetModulation(fLengthRatio)


__all__ = ["LatticePipe"]



from __future__ import annotations

import ctypes
from dataclasses import dataclass
from typing import Iterable

from _common.types import Vector3Like, as_tuple3


class Vec2(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float), ("y", ctypes.c_float)]


class Vec3(ctypes.Structure):
    _fields_ = [("x", ctypes.c_float), ("y", ctypes.c_float), ("z", ctypes.c_float)]


class Triangle(ctypes.Structure):
    _fields_ = [("a", ctypes.c_int), ("b", ctypes.c_int), ("c", ctypes.c_int)]


class ColorFloat(ctypes.Structure):
    _fields_ = [
        ("r", ctypes.c_float),
        ("g", ctypes.c_float),
        ("b", ctypes.c_float),
        ("a", ctypes.c_float),
    ]


class BBox3(ctypes.Structure):
    _fields_ = [("min", Vec3), ("max", Vec3)]


_BBox3C = BBox3


class Mat4(ctypes.Structure):
    _fields_ = [(f"m{i}", ctypes.c_float) for i in range(16)]

@dataclass(frozen=True)
class VoxelDimensions:
    x_origin: int
    y_origin: int
    z_origin: int
    x_size: int
    y_size: int
    z_size: int


def as_vec3(value: Vector3Like | Vec3) -> Vec3:
    if isinstance(value, Vec3):
        return value
    x, y, z = as_tuple3(value)
    return Vec3(x, y, z)


def as_bbox3(min_xyz: Vector3Like, max_xyz: Vector3Like) -> BBox3:
    return BBox3(as_vec3(min_xyz), as_vec3(max_xyz))


def vec3_tuple(v: Vec3) -> tuple[float, float, float]:
    return (float(v.x), float(v.y), float(v.z))


def iter_vec3(values: Iterable[Vector3Like]) -> Iterable[Vec3]:
    for value in values:
        yield as_vec3(value)



class PicoGKError(RuntimeError):
    """Base exception for picogk-python."""


class PicoGKLoadError(PicoGKError):
    """Raised when the native runtime cannot be loaded."""


class PicoGKDisposedError(PicoGKError):
    """Raised when a disposed native wrapper is used."""


class PicoGKInvalidHandleError(PicoGKError):
    """Raised when the runtime returns a null/invalid handle."""
